﻿using System;
namespace ProbitClassification
{
  class ProbitProgram
  {
    static void Main(string[] args)
    {
      Console.WriteLine("\nBegin Probit Binary Classification demo");
      Console.WriteLine("Goal is to predict death (0 = false, 1 = true)");

      double[][] data = new double[30][];
      data[0] = new double[] { 48, +1, 4.40, 0 };
      data[1] = new double[] { 60, -1, 7.89, 1 };
      data[2] = new double[] { 51, -1, 3.48, 0 };
      data[3] = new double[] { 66, -1, 8.41, 1 };
      data[4] = new double[] { 40, +1, 3.05, 0 };
      data[5] = new double[] { 44, +1, 4.56, 0 };
      data[6] = new double[] { 80, -1, 6.91, 1 };
      data[7] = new double[] { 52, -1, 5.69, 0 };
      data[8] = new double[] { 56, -1, 4.01, 0 };
      data[9] = new double[] { 55, -1, 4.48, 0 };
      data[10] = new double[] { 72, +1, 5.97, 0 };
      data[11] = new double[] { 57, -1, 6.71, 1 };
      data[12] = new double[] { 50, -1, 6.40, 0 };
      data[13] = new double[] { 80, -1, 6.67, 1 };
      data[14] = new double[] { 69, +1, 5.79, 0 };
      data[15] = new double[] { 39, -1, 5.42, 0 };
      data[16] = new double[] { 68, -1, 7.61, 1 };
      data[17] = new double[] { 47, +1, 3.24, 0 };
      data[18] = new double[] { 45, +1, 4.29, 0 };
      data[19] = new double[] { 79, +1, 7.44, 1 };
      data[20] = new double[] { 44, -1, 2.55, 0 };
      data[21] = new double[] { 52, +1, 3.71, 0 };
      data[22] = new double[] { 80, +1, 7.56, 1 };
      data[23] = new double[] { 76, -1, 7.80, 1 };
      data[24] = new double[] { 51, -1, 5.94, 0 };
      data[25] = new double[] { 46, +1, 5.52, 0 };
      data[26] = new double[] { 48, -1, 3.25, 0 };
      data[27] = new double[] { 58, +1, 4.71, 0 };
      data[28] = new double[] { 44, +1, 2.52, 0 };
      data[29] = new double[] { 68, -1, 8.38, 1 };

      Console.WriteLine("\nRaw data: \n");
      Console.WriteLine("       Age       Sex      Kidney   Died");
      Console.WriteLine("=======================================");
      ShowData(data, 5, 2, true);

      Console.WriteLine("Normalizing age and kidney data");
      int[] columns = new int[] { 0, 2 };
      double[][] means = Normalize(data, columns); // normalize, save means & stdDevs
      Console.WriteLine("Done");

      Console.WriteLine("\nNormalized data: \n");
      ShowData(data, 5, 2, true);

      Console.WriteLine("Creating train (80%) and test (20%) matrices");
      double[][] trainData;
      double[][] testData;
      MakeTrainTest(data, 0, out trainData, out testData);
      Console.WriteLine("Done");

      Console.WriteLine("\nNormalized training data: \n");
      ShowData(trainData, 3, 2, true);

      //Console.WriteLine("\nFirst 3 rows and last row of normalized test data: \n");
      //ShowData(testData, 3, 2, true);

      int numFeatures = 3; // number of x values (age, sex, kidney)
      Console.WriteLine("Creating probit binary classifier");
      ProbitClassifier pc = new ProbitClassifier(numFeatures); //

      int maxEpochs = 100; // 100 gives a representative demo
      Console.WriteLine("Setting maxEpochs = " + maxEpochs);
      Console.WriteLine("Starting training");
      //double[] bestWeights = pc.Train(trainData, maxEpochs);
      double[] bestWeights = pc.Train(trainData, maxEpochs, 0);
      Console.WriteLine("Training complete");

      Console.WriteLine("\nBest weights found:");
      ShowVector(bestWeights, 4, true);

      double trainAccuracy = pc.Accuracy(trainData, bestWeights);
      Console.WriteLine("Prediction accuracy on training data = " + trainAccuracy.ToString("F4"));

      double testAccuracy = pc.Accuracy(testData, bestWeights);
      Console.WriteLine("Prediction accuracy on test data = " + testAccuracy.ToString("F4"));

      //double[] unknownNormalized = new double[] { 0.25, -1.0, 0.50 }; // slightly older, male, higher kidney
      //int died = pc.ComputeDependent(unknownNormalized, bestWeights);
      //if (died == 0)
      //  Console.WriteLine("Predict survive");
      //else if (died == 1)
      //  Console.WriteLine("Predict die");

      Console.WriteLine("\nEnd probit binary classification demo\n");
      Console.ReadLine();
    }

    static double[][] Normalize(double[][] rawData, int[] columns)
    {
      // return means and sdtDevs of all columns for later use
      int numRows = rawData.Length;
      int numCols = rawData[0].Length;

      double[][] result = new double[2][]; // [0] = mean, [1] = stdDev
      for (int i = 0; i < 2; ++i)
        result[i] = new double[numCols]; // 

      for (int c = 0; c < numCols; ++c)
      {
        // means all cols
        double sum = 0.0;
        for (int r = 0; r < numRows; ++r)
          sum += rawData[r][c];
        double mean = sum / numRows;
        result[0][c] = mean; // save

        // stdDevs all cols
        double sumSquares = 0.0;
        for (int r = 0; r < numRows; ++r)
          sumSquares += (rawData[r][c] - mean) * (rawData[r][c] - mean);
        double stdDev = Math.Sqrt(sumSquares / numRows);
        result[1][c] = stdDev;
      }

      // normalize
      for (int c = 0; c < columns.Length; ++c)
      {
        int j = columns[c]; // column to normalize
        double mean = result[0][j]; // mean of the col
        double stdDev = result[1][j];
        for (int i = 0; i < numRows; ++i)
          rawData[i][j] = (rawData[i][j] - mean) / stdDev;
      }
      return result;
    }

    static void Normalize(double[][] rawData, int[] columns, double[][] means)
    {
      // normalize columns using supplied means and standard devs
      int numRows = rawData.Length;
      for (int c = 0; c < columns.Length; ++c) // each specified col
      {
        int j = columns[c]; // column to normalize
        double mean = means[0][j];
        double stdDev = means[1][j];
        for (int i = 0; i < numRows; ++i) // each row
          rawData[i][j] = (rawData[i][j] - mean) / stdDev;
      }
    }

    static void MakeTrainTest(double[][] allData, int seed, out double[][] trainData, out double[][] testData)
    {
      Random rnd = new Random(seed);
      int totRows = allData.Length;
      int numTrainRows = (int)(totRows * 0.80); // 80% hard-coded
      int numTestRows = totRows - numTrainRows;
      trainData = new double[numTrainRows][];
      testData = new double[numTestRows][];

      double[][] copy = new double[allData.Length][]; // ref copy of all data
      for (int i = 0; i < copy.Length; ++i)
        copy[i] = allData[i];

      for (int i = 0; i < copy.Length; ++i) // scramble order
      {
        int r = rnd.Next(i, copy.Length);
        double[] tmp = copy[r];
        copy[r] = copy[i];
        copy[i] = tmp;
      }
      for (int i = 0; i < numTrainRows; ++i)
        trainData[i] = copy[i];

      for (int i = 0; i < numTestRows; ++i)
        testData[i] = copy[i + numTrainRows];
    } // MakeTrainTest


    static void ShowData(double[][] data, int numRows, int decimals, bool indices)
    {
      for (int i = 0; i < numRows; ++i)
      {
        if (indices == true)
          Console.Write("[" + i.ToString().PadLeft(2) + "]  ");
        for (int j = 0; j < data[i].Length; ++j)
        {
          double v = data[i][j];
          if (v >= 0.0)
            Console.Write(" "); // '+'
          Console.Write(v.ToString("F" + decimals) + "    ");
        }
        Console.WriteLine("");
      }
      Console.WriteLine(". . .");
      int lastRow = data.Length - 1;
      if (indices == true)
        Console.Write("[" + lastRow.ToString().PadLeft(2) + "]  ");
      for (int j = 0; j < data[lastRow].Length; ++j)
      {
        double v = data[lastRow][j];
        if (v >= 0.0)
          Console.Write(" "); // '+'
        Console.Write(v.ToString("F" + decimals) + "    ");
      }
      Console.WriteLine("\n");
    }

    static void ShowVector(double[] vector, int decimals, bool newLine)
    {
      for (int i = 0; i < vector.Length; ++i)
        Console.Write(vector[i].ToString("F" + decimals) + " ");
      Console.WriteLine("");
      if (newLine == true)
        Console.WriteLine("");
    }

    static double CumDensity(double z)
    {
      // aka Phi(z)
      // Abramowitz and Stegun eq. 7.1.26
      double p = 0.3275911;
      double a1 = 0.254829592;
      double a2 = -0.284496736;
      double a3 = 1.421413741;
      double a4 = -1.453152027;
      double a5 = 1.061405429;

      int sign;
      if (z < 0.0)
        sign = -1;
      else
        sign = 1;

      double x = Math.Abs(z) / Math.Sqrt(2.0);
      double t = 1.0 / (1.0 + p * x);
      double erf = 1.0 - (((((a5 * t + a4) * t) + a3) * t + a2) * t + a1) * t * Math.Exp(-x * x);
      return 0.5 * (1.0 + (sign * erf));
    }


  } // Program

  public class ProbitClassifier
  {
    private int numFeatures; // number of indpendent variables
    private double[] weights; // b0 = constant
    private Random rnd;

    public ProbitClassifier(int numFeatures)
    {
      this.numFeatures = numFeatures;
      this.weights = new double[numFeatures + 1]; // [0] = b0 constant
      //rnd = new Random(11); // gives representative demo with maxEpochs = 100
    }

    public double[] Train(double[][] trainData, int maxEpochs, int seed)
    {
      // sort 3 solutions (small error to large)
      // compute centroid 
      // if expanded better than worst, replace
      // else if reflected better than worst, replace
      // else if contracted better than worst, replace
      // else if random better than worst, replace
      // else shrink

      this.rnd = new Random(seed);

      Solution[] solutions = new Solution[3]; // best, worst, other

      // initialize to random values
      for (int i = 0; i < 3; ++i)
      {
        solutions[i] = new Solution(numFeatures);
        solutions[i].weights = RandomSolution();
        solutions[i].error = Error(trainData, solutions[i].weights);
      }

      int epoch = 0;
      while (epoch < maxEpochs)
      {
        ++epoch;
        Array.Sort(solutions); // [0] = best, [1] = other, [2] = worst
        double[] best = solutions[0].weights; // smallest error
        double[] other = solutions[1].weights;
        double[] worst = solutions[2].weights; // largest error
        // consider exit if  best == worst

        double[] centroid = Centroid(other, best); // an average

        double[] expanded = Expanded(centroid, worst);
        double expandedError = Error(trainData, expanded); 
        if (expandedError < solutions[2].error) // expanded better than worst?
        {
          Array.Copy(expanded, solutions[2].weights, numFeatures + 1);
          solutions[2].error = expandedError;
          continue;
        }

        double[] reflected = Reflected(centroid, worst);
        double reflectedError = Error(trainData, reflected); 
        if (reflectedError < solutions[2].error) // relected better than worst?
        {
          Array.Copy(reflected, solutions[2].weights, numFeatures + 1);
          solutions[2].error = reflectedError;
          continue;
        }

        double[] contracted = Contracted(centroid, worst);
        double contractedError = Error(trainData, contracted); 
        if (contractedError < solutions[2].error)
        {
          Array.Copy(contracted, solutions[2].weights, numFeatures + 1);
          solutions[2].error = contractedError;
          continue;
        }

        // consder only every 10 epochs
        double[] randomSol = RandomSolution();
        double randomSolError = Error(trainData, randomSol);
        if (randomSolError < solutions[2].error)
        {
          Array.Copy(randomSol, solutions[2].weights, numFeatures + 1);
          solutions[2].error = randomSolError;
          continue;
        }

        // couldn't find a replacement for worst so shrink
        // worst -> towards best
        for (int j = 0; j < numFeatures + 1; ++j)
        {
          solutions[2].weights[j] = (solutions[2].weights[j] + solutions[0].weights[j]) / 2.0;
        }
        solutions[2].error = Error(trainData, solutions[2].weights);

        // other -> towards best
        for (int j = 0; j < numFeatures + 1; ++j)
        {
          solutions[1].weights[j] = (solutions[1].weights[j] + solutions[0].weights[j]) / 2.0;
        }
        solutions[1].error = Error(trainData, solutions[1].weights);

      } // while

      // copy best weights found
      Array.Copy(solutions[0].weights, this.weights, this.numFeatures + 1);
      return this.weights;
    }

    private double[] Centroid(double[] other, double[] best) // centroid doesn't use worst
    {
      double[] result = new double[this.numFeatures + 1];
      for (int i = 0; i < result.Length; ++i)
        result[i] = (other[i] + best[i]) / 2.0;
      return result;
    }

    private double[] Reflected(double[] centroid, double[] worst)
    {
      double alpha = 1.0; // how far from centroid
      double[] result = new double[this.numFeatures + 1];
      for (int i = 0; i < result.Length; ++i)
        result[i] = centroid[i] + (alpha * (centroid[i] - worst[i]));
      //result[i] = ((1 + alpha) * centroid[i]) - (alpha * worst[i]);
      return result;
    }

    private double[] Expanded(double[] centroid, double[] worst)
    {
      double gamma = 2.0; // how far
      double[] result = new double[this.numFeatures + 1];
      for (int i = 0; i < result.Length; ++i)
        //result[i] = (gamma * reflected[i]) + ((1 - gamma) * centroid[i]);
        result[i] = centroid[i] + (gamma * (centroid[i] - worst[i]));
      return result;
    }

    private double[] Contracted(double[] centroid, double[] worst)
    {
      double rho = -0.5;
      double[] result = new double[this.numFeatures + 1];
      for (int i = 0; i < result.Length; ++i)
        //result[i] = (beta * worst[i]) + ((1 - beta) * centroid[i]);
        result[i] = centroid[i] + (rho * (centroid[i] - worst[i]));
      return result;
    }

    private double[] RandomSolution()
    {
      double[] result = new double[this.numFeatures + 1];
      double lo = -10.0;
      double hi = 10.0;
      for (int i = 0; i < result.Length; ++i)
        result[i] = (hi - lo) * rnd.NextDouble() + lo;
      return result;
    }

    private double Error(double[][] trainData, double[] weights)
    {
      // mean squared error using supplied weights
      int yIndex = trainData[0].Length - 1; // y-value (0/1) is last column
      double sumSquaredError = 0.0;
      for (int i = 0; i < trainData.Length; ++i) // each data
      {
        double[] xData = new double[numFeatures]; // no Y
        Array.Copy(trainData[i], xData, numFeatures);
        double computed = ComputeOutput(xData, weights); // ex: 0.383
        double desired = trainData[i][yIndex]; // ex: 0.0 or 1.0
        sumSquaredError += (computed - desired) * (computed - desired);
      }
      return sumSquaredError / trainData.Length;
    }

    public void SetWeights(double[] weights)
    {
      Array.Copy(weights, this.weights, weights.Length);
    }

    public double[] GetWeights()
    {
      double[] result = new double[weights.Length];
      Array.Copy(this.weights, result, weights.Length);
      return result;
    }

    public double ComputeOutput(double[] dataItem, double[] weights)
    {
      //if (xData.Length != weights.Length - 1) // tricky!
      //  throw new Exception("Bad xData in ComputeOutputs");

      double z = 0.0;

      z += weights[0]; // constant
      for (int i = 0; i < weights.Length - 1; ++i) // data might include Y
        z += (weights[i + 1] * dataItem[i]); // skip first weight
      return CumDensity(z);
    }

    private static double CumDensity(double z) // probit classification
    {
      // Abramowitz and Stegun eq. 7.1.26
      double p = 0.3275911;
      double a1 = 0.254829592;
      double a2 = -0.284496736;
      double a3 = 1.421413741;
      double a4 = -1.453152027;
      double a5 = 1.061405429;

      int sign;
      if (z < 0.0)
        sign = -1;
      else
        sign = 1;

      double x = Math.Abs(z) / Math.Sqrt(2.0);
      double t = 1.0 / (1.0 + p * x);
      double erf = 1.0 - (((((a5 * t + a4) * t) + a3) * t + a2) * t + a1) * t * Math.Exp(-x * x);
      return 0.5 * (1.0 + sign * erf);
    }



    public int ComputeDependent(double[] dataItem, double[] weights)
    {
      double sum = ComputeOutput(dataItem, weights);

      if (sum <= 0.5)
        return 0;
      else
        return 1;
    }

    public double Accuracy(double[][] trainData, double[] weights)
    {
      int numCorrect = 0;
      int numWrong = 0;
      int yIndex = trainData[0].Length - 1;
      for (int i = 0; i < trainData.Length; ++i)
      {
        double[] xData = new double[numFeatures]; // don't want Y
        Array.Copy(trainData[i], xData, numFeatures);
        double computed = ComputeDependent(xData, this.weights); // 0.0 or 1.0
        double desired = trainData[i][yIndex]; // 0.0 or 1.0

        //Console.WriteLine(i + " " + "Computed = " + computed + " desired = " + desired);
        if (computed == desired)
          ++numCorrect;
        else
          ++numWrong;
      }
      return (numCorrect * 1.0) / (numWrong + numCorrect);
    }


    private class Solution : IComparable<Solution>
    {
      public double[] weights; // a potential solution (to minimize LR bi error
      public double error;    // value of function to minimize at vector[]

      public Solution(int numFeatures)
      {
        this.weights = new double[numFeatures + 1]; // problem dim + constant b0
        this.error = 0.0;
      }

      public int CompareTo(Solution other) // low-to-high error
      {
        if (this.error < other.error)
          return -1;
        else if (this.error > other.error)
          return 1;
        else
          return 0;
      }
    } // Solution   

  } // ProbitClassifier


} // ns
